from duckduckgo_search import DDGS
from pylmkit.utils.data_utils import Document
from pylmkit.core.base import BaseKnowledgeBase


class WebSearch(DDGS, BaseKnowledgeBase):
    def __init__(
            self,
            topk=5,
            backend="api",
            region="wt-wt",
            timelimit=None,
            safesearch="moderate",
            init_documents=None,
            timeout=10,
            headers=None,
            proxies=None
    ):
        DDGS.__init__(
            self,
            timeout=timeout,
            headers=headers,
            proxies=proxies
        )
        BaseKnowledgeBase.__init__(self, init_documents=init_documents)
        self.topk = int(topk)
        self.backend = backend
        self.region = region
        self.timelimit = timelimit
        self.safesearch = safesearch

    def get(self, keyword):
        if keyword:
            search_gen = super().text(keywords=keyword,
                                      backend=self.backend,
                                      region=self.region,
                                      max_results=self.topk,
                                      timelimit=self.timelimit,
                                      safesearch=self.safesearch
                                      )
            for i, page in enumerate(list(search_gen)):
                if page:
                    self.documents.append(Document(
                        page_content=page['body'],
                        metadata={"source": page.get('href', ''), "title": page.get('title', '')},
                        type="Web search"
                    ))
            return self.documents
        else:
            raise Exception("Keyword cannot be empty")





import logging


class BaseLogging(object):
    def __init__(self):
        # logging setting
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)s %(name)s %(levelname)s %(message)s',
                            datefmt='%Y-%m-%d %H:%M:%S',  # 定义日期格式
                            )

        self.logger = logging.getLogger(__file__)  # 当前.py文件路径





from pylmkit.core import BaseMemory


class MemoryHistoryLength(BaseMemory):
    """

    例子
    >>> from pylmkit.memory import MemoryHistoryLength
    >>> mhl = MemoryHistoryLength(memory_length=100)
    >>> mhl.add(role="user", content="中国首都在哪个城市？")
    >>> mhl.add(role="ai", content="北京")
    >>> mhl.add(role="user", content="广州属于哪个省？")
    >>> mhl.add(role="ai", content="广东")
    >>> print(mhl.get())
    >>> print(mhl.memory_length)
    """

    def __init__(self, memory_length=1000, init_memory=None, streamlit_web=False):
        super().__init__(init_memory=init_memory, streamlit_web=streamlit_web)
        self.memory_length = int(memory_length)

    def get(self):
        history_memory = super()._get(mode="string")
        return history_memory[-self.memory_length:]

    def origin_memory(self):
        return self.memory_messages


class MemoryConversationsNumber(BaseMemory):
    """
    >>> from pylmkit.memory import MemoryConversationsNumber
    >>> mcn = MemoryConversationsNumber(number=1)
    >>> mcn.add(role="user", content="中国首都在哪个城市？")
    >>> mcn.add(role="ai", content="北京")
    >>> mcn.add(role="user", content="广州市属于中国哪个省？")
    >>> mcn.add(role="ai", content="广东省")
    >>> print(mcn.get())
    """

    def __init__(self, number=5, init_memory=None, streamlit_web=False):
        super().__init__(init_memory=init_memory, streamlit_web=streamlit_web)
        self.number = int(number)

    def get(self):
        conversations_memory = super()._get(mode="message")
        return conversations_memory[-self.number * 2:]

    def origin_memory(self):
        return self.memory_messages


class MemorySummary(BaseMemory):
    def __init__(self, llm_model, history_memory_length=1000, prompt_template: str = "", init_memory=None, streamlit_web=False):
        super().__init__(init_memory=init_memory, streamlit_web=streamlit_web)
        self.model = llm_model
        self.history_memory_length = int(history_memory_length)
        self.prompt_template = str(prompt_template)
        if not prompt_template:
            self.prompt_template = "根据下面历史对话内容提取摘要\n"

    def get(self):
        history_memory = super()._get(mode="string")[-self.history_memory_length:]
        query = self.prompt_template + history_memory
        summary_memory = self.model.invoke(query)
        return summary_memory

    def origin_memory(self):
        return self.memory_messages





"""
Reuse existing LM APIs from langchain.
Additionally, add non-existent APIs.
And further standardize the API name
"""
from typing import Any


def _import_ChatQianfan() -> Any:
    from langchain.llms import QianfanLLMEndpoint
    from langchain.chat_models import QianfanChatEndpoint

    return QianfanLLMEndpoint


def _import_ChatOpenAI() -> Any:
    from langchain.chat_models import ChatOpenAI

    return ChatOpenAI


def _import_ChatTongyi() -> Any:
    from langchain.chat_models.tongyi import ChatTongyi

    return ChatTongyi


def _import_LocalLLMModel() -> Any:
    from pylmkit.llms._huggingface_llm import LocalLLMModel

    return LocalLLMModel


def _import_ChatHunyuan() -> Any:
    from langchain.chat_models import ChatHunyuan

    return ChatHunyuan


def _import_ChatBaichuan() -> Any:
    from langchain.chat_models import ChatBaichuan

    return ChatBaichuan


def _import_ChatZhipu() -> Any:
    from pylmkit.llms._zhipu import ChatZhipu

    return ChatZhipu


def _import_ChatSpark() -> Any:
    from pylmkit.llms._spark import ChatSpark

    return ChatSpark


def _import_EmbeddingsQianfan() -> Any:
    from langchain.embeddings import QianfanEmbeddingsEndpoint

    return QianfanEmbeddingsEndpoint


def _import_EmbeddingsHuggingFace() -> Any:
    from langchain.embeddings import HuggingFaceEmbeddings

    return HuggingFaceEmbeddings


def _import_EmbeddingsHuggingFaceInstruct() -> Any:
    from langchain.embeddings import HuggingFaceInstructEmbeddings

    return HuggingFaceInstructEmbeddings


def _import_EmbeddingsHuggingFaceBge() -> Any:
    from langchain.embeddings import HuggingFaceBgeEmbeddings

    return HuggingFaceBgeEmbeddings


def _import_EmbeddingsOpenAI() -> Any:
    from langchain.embeddings import OpenAIEmbeddings

    return OpenAIEmbeddings


def __getattr__(name: str) -> Any:
    if name == "ChatQianfan":
        return _import_ChatQianfan()
    elif name == "ChatOpenAI":
        return _import_ChatOpenAI()
    elif name == "ChatHunyuan":
        return _import_ChatHunyuan()
    elif name == "ChatBaichuan":
        return _import_ChatBaichuan()
    elif name == "ChatZhipu":
        return _import_ChatZhipu()
    elif name == "ChatSpark":
        return _import_ChatSpark()
    elif name == "ChatTongyi":
        return _import_ChatTongyi()
    elif name == "LocalLLMModel":
        return _import_LocalLLMModel()
    elif name == "EmbeddingsQianfan":
        return _import_EmbeddingsQianfan()
    elif name == "EmbeddingsHuggingFace":
        return _import_EmbeddingsHuggingFace()
    elif name == "EmbeddingsHuggingFaceInstruct":
        return _import_EmbeddingsHuggingFaceInstruct()
    elif name == "EmbeddingsHuggingFaceBge":
        return _import_EmbeddingsHuggingFaceBge()
    elif name == "EmbeddingsOpenAI":
        return _import_EmbeddingsOpenAI()
    else:
        raise AttributeError(f"Could not find: {name}")


__all__ = [
    "ChatQianfan",
    "ChatBaichuan",
    "ChatHunyuan",
    "ChatOpenAI",
    "ChatTongyi",
    "ChatSpark",
    "ChatZhipu",
    "LocalLLMModel",
    "EmbeddingsQianfan",
    "EmbeddingsHuggingFace",
    "EmbeddingsHuggingFaceInstruct",
    "EmbeddingsHuggingFaceBge",
    "EmbeddingsOpenAI",

]







# main.py
from pylmkit import BaseWebUI
from dotenv import load_dotenv
from pylmkit.app import RolePlay
from pylmkit.llms import ChatOpenAI
from pylmkit.memory import MemoryHistoryLength
from pylmkit.llms import ChatQianfan


load_dotenv()
web = BaseWebUI(language='zh')  # 中文网站


model = ChatQianfan(model="ERNIE-Bot-turbo")
memory = MemoryHistoryLength(memory_length=web.param(label="记忆长度", type='int', value=500),  # 添加页面交互参数
                             streamlit_web=True
                            )
role_template = "{memory}\n 请为我推荐{query}的{topic}"
rp = RolePlay(
    role_template=role_template,  # 角色模板
    llm_model=model,  # 大语言模型
    memory=memory,  # 记忆
    # online_search_kwargs={},
    online_search_kwargs={'topk': 2, 'timeout': 20},  # 搜索引擎配置，不开启则可以设置为 online_search_kwargs={}
    return_language="中文"
)

web.run(
    obj=rp.invoke,
    input_param=[{"name": "query", "label": "地点", "type": "chat"},
                 {"name": "topic", "label": "主题", "type": "text"},
                 ],
    output_param=[{'label': '结果', 'name': 'response', 'type': 'chat'},
                  {'label': '参考', 'name': 'refer', 'type': 'refer'}
                  ]
)





from langchain.document_loaders import UnstructuredFileLoader
from langchain.document_loaders import TextLoader
from langchain.document_loaders import DirectoryLoader


# loader = UnstructuredFileLoader("../document_test/aaa.txt", encodings="utf-8")
# docs = loader.load()
# print(len(docs), docs)
#
# with open("../document_test/doc.json", 'r', encoding='utf-8') as fp:
#     line = fp.read()
#     print(line)

loader = DirectoryLoader('../document_test',
                         glob="**/*",
                         silent_errors=True,
                         )
docs = loader.load()
print(len(docs), len(docs[4].page_content))





from pylmkit.utils.data_utils import message_as_string
from pylmkit.utils.data_utils import document_as_string
from pylmkit.utils.data_utils import document_as_refer
from pylmkit.core.prompt import return_language


class RolePlay(object):
    from pylmkit.utils.logging import BaseLogging
    log = BaseLogging().logger

    def __init__(self,
                 role_template="",
                 llm_model=None,
                 memory=None,
                 online_search_kwargs={},
                 return_language="English",
                 **kwargs
                 ):
        self.role_template = role_template
        self.model = llm_model
        self.return_language = return_language
        self.memory = memory
        if online_search_kwargs:
            from pylmkit.tools.search import WebSearch
            self.search = WebSearch(**online_search_kwargs)
        else:
            self.search = None

    def invoke(self, query: str, ra_documents=[], **kwargs):
        prompt, refer = self._invoke(query, ra_documents=ra_documents, **kwargs)
        response = self.model.invoke(prompt)
        response = [response if isinstance(response, str) else response.content][0]
        if self.memory:
            if not refer:
                self.memory.add(role='ai', content=response)  # add ai output
            else:
                self.memory.add(role='ai', content=response, refer=refer)
        return response, refer

    def stream(self, query: str, ra_documents=[], **kwargs):
        prompt, refer = self._invoke(query, ra_documents=ra_documents, **kwargs)
        response = self.model.stream(prompt)
        if self.memory:
            if not refer:
                self.memory.add(role='ai',
                                content="".join([i.content if not isinstance(i, str) else i for i in response])
                                )  # add ai output
            else:
                self.memory.add(role='ai',
                                content="".join([i.content if not isinstance(i, str) else i for i in response]),
                                refer=refer
                                )  # add ai output
        return response, refer

    def return_memory(self):
        return self.memory.memory_messages

    def clear_memory(self):
        self.memory.clear()

    def _invoke(self, query, ra_documents=[], **kwargs):
        # search
        search_content = ""
        search_documents = []
        if self.search and query:
            try:
                search_documents = self.search.get(keyword=query)
                search_content = document_as_string(documents=search_documents)
            except Exception as e:
                # print(f">>> Online search except: {e}")
                self.log.error(f"{e}")
        # ra base
        ra_content = ""
        if ra_documents:
            ra_content = document_as_string(documents=ra_documents)
        # refer
        refer = []
        refer.extend(search_documents)
        refer.extend(ra_documents)
        refer = document_as_refer(refer)
        # memory
        memory_content = ""
        if self.memory and query:
            memory_content = self.memory.get()
            if isinstance(memory_content, list):
                memory_content = message_as_string(memory_content)
            self.memory.add(content=query, role='human')  # add user input
        # prompt
        try:
            prompt_kwargs = {"search": search_content,
                             'memory': memory_content,
                             "query": query,
                             "ra": ra_content,
                             }
            prompt_kwargs.update(kwargs)
            prompt = self.role_template.format(**prompt_kwargs)
        except Exception as e:
            prompt = query
            # print(f"role template except: {e}")
            self.log.error(f"Role template except: {e}")
        # return language
        if self.return_language:
            prompt += return_language(language=self.return_language)
        return prompt, refer




from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='pylmkit',  # 对外我们模块的名字
    author='52phm',  # 作者
    author_email='2374521450@qq.com',  # 作者邮箱
    version='0.0.16',  # 版本号
    description='pylmkit: Help users quickly build practical large model applications!',  # 项目摘要
    long_description=long_description,  # 项目介绍
    long_description_content_type="text/markdown",
    url="https://github.com/52phm/pylmkit",  # 程序主页地址
    packages=find_packages(),  # 要发布的模块
    license="Apache License-2.0",  # 协议（请注意按实际情况选择不同的协议），可选
    # 此处添加任意多个项目链接
    project_urls={
        # 英文关键字是内置好的关键字，有对应的logo
        "Homepage": "http://app.pylmkit.cn",
        "Documentation": "http://en.pylmkit.cn",
        # 也可以任意命名，默认统一logo
        "应用主页": "http://app.pylmkit.cn",
        "中文文档": "http://zh.pylmkit.cn",
    },
    install_requires=[
        'streamlit',
        'pyyaml',
        "dashscope",
        "zhipuai",
        "openai",
        "python-dotenv",
        'langchain==0.0.325',
        'qianfan',
        'duckduckgo_search',
        "websocket==0.2.1",
        "websocket-client==1.6.3",
        # "unstructured",
        # "pdf2image",
        # "pdfminer.six",
        # "sentence-transformers",
        # "faiss-cpu",
        # "chromadb",
    ]
)
